@@ -24,6 +24,7 @@
 #pragma once
 
 #include <osmocom/core/linuxlist.h>
+#include <osmocom/core/hashtable.h>
 #include <osmocom/core/select.h>
 #include <osmocom/core/logging.h>
 
@@ -55,6 +56,8 @@ struct upf_gtp_dev {
 
 	/* list of struct upf_gtp_tunend */
 	struct llist_head tunnels;
+	/* hashtable of (struct upf_gtp_tunen) with key desc.access.local.teid */
+	DECLARE_HASHTABLE(tunnels_by_local_f_teid, 10);
 };
 
 /* Description of a GTP encapsulation / decapsulation.
